/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include <gdam/gdamchannel.h>
#include <gdam/gdam-protocol.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <gdam/autoconf.h>
#if HAVE_LIBREADLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif
#include <gdam/gdamarg.h>
#include "linetokenizer.h"
#include <unistd.h>

typedef struct _CliInfo CliInfo;

struct _CliInfo {
	GdamChannel* channel;
	int device_id;
	int mixer_id;
	unsigned interactive : 1;
};

#if !HAVE_LIBREADLINE
/* Emulate libreadline's fabulous design. */
static char* readline(const char* prompt)
{
	static char buf[2048];
	fprintf(stderr, "> ");
	return fgets(buf, sizeof(buf), stdin) ? buf : NULL;
}
static void add_history(const char* line)
{
}
#endif


/* Map from strings to other strings. */
static GHashTable* variables = NULL;


/* A function to process a single type of command, and a hashtable
 * mapping names to handlers. */
typedef gboolean (*CommandHandler)(LineTokenizer* tokenizer,
                                   CliInfo*       info,
				   GdamRequest**  request);

/* Map from string (command name) => CommandHandler. */
static GHashTable* command_handlers = NULL;

void cli_help(const char* topic);

static void initial_empty_hashtables()
{
	variables = g_hash_table_new(g_str_hash, g_str_equal);
	command_handlers = g_hash_table_new(g_str_hash, g_str_equal);
}

/*
 * 
 *  ___  ___  _   _ _ __ ___ ___
 * / __|/ _ \| | | | '__/ __/ _ \
 * \__ \ (_) | |_| | | | (_|  __/
 * |___/\___/ \__,_|_|  \___\___|
 * 
 * Create a sound source on the server side and assign its
 * id to a variable.
 */
typedef struct _SourceCommandInfo SourceCommandInfo;
struct _SourceCommandInfo {
	char* variable_name;
	CliInfo* cli_info;
};

static void handle_source_command_success(int source_id, gpointer ptr_info)
{
	SourceCommandInfo* info = (SourceCommandInfo*)ptr_info;

	/* Whether to free the variable name.
	 * This should be done if the variable is already assigned.
	 * (BTW g_hash_table_insert will use the old key in case of 
	 *      a collision.) */
	gboolean must_free_variable_name = FALSE;


	if (info->cli_info->interactive)
		fprintf(stderr, "variable `%s' assigned source id %d.\n",
			info->variable_name, source_id);

	if (g_hash_table_lookup(variables, info->variable_name))
		must_free_variable_name = TRUE;

	g_hash_table_insert(variables,
			    info->variable_name,
			    g_strdup_printf("%d", source_id));

	if (must_free_variable_name)
		g_free(info->variable_name);

	g_free(info);
}

static void handle_source_command_failure(gpointer ptr_info)
{
	SourceCommandInfo* info = (SourceCommandInfo*)ptr_info;
	if (info->cli_info->interactive)
		fprintf(stderr, "variable `%s' not assigned as source "
				"was not created.\n", info->variable_name);
	g_free(info->variable_name);
	g_free(info);
}


static gboolean handle_source_command(LineTokenizer* tokenizer,
                                      CliInfo*       info,
				      GdamRequest**  request)
{
	char* name = line_tokenizer_get_word(tokenizer);
	char* type = line_tokenizer_get_word(tokenizer);
	int num_args = 0, max_args = 16;
	GdamArg* args = g_new(GdamArg, max_args);
	SourceCommandInfo* command_info = g_new(SourceCommandInfo, 1);
	command_info->variable_name = name;
	command_info->cli_info = info;
	while (1) {
		if (num_args == max_args) {
			max_args *= 2;
			args = g_renew(GdamArg, args, max_args);
		}
		if (!line_tokenizer_tokenize(tokenizer,
					     args + num_args)) {
			break;
		}
		num_args++;
	}
	*request = gdam_request_create_source_v(
		type,
		num_args,
		args,
		handle_source_command_success,
		handle_source_command_failure,
		command_info
	);
	gdam_arg_free_array(num_args, args);
	g_free(type);
	return *request ? TRUE : FALSE;
}

/*
 *           _
 *  ___  ___| |_
 * / __|/ _ \ __|
 * \__ \  __/ |_
 * |___/\___|\__|
 * Set arguments for a source.
 */
static void handle_set_command_success(gpointer unused)
{
	gdam_debug("set command succeeded");
}
static void handle_set_command_failure(gpointer unused)
{
	gdam_debug("set command failed");
}
static gboolean handle_set_command(LineTokenizer* tokenizer,
                                   CliInfo*       info,
				   GdamRequest**  request)
{
	char* id = line_tokenizer_get_word(tokenizer);
	int source_id;
	int num_args = 0, max_args = 16;
	GdamArg* args = g_new(GdamArg, max_args);
	while (1) {
		if (num_args == max_args) {
			max_args *= 2;
			args = g_renew(GdamArg, args, max_args);
		}
		if (!line_tokenizer_tokenize(tokenizer,
					     args + num_args)) {
			break;
		}
		num_args++;
	}
	source_id = (int) strtol(id, NULL, 10);
	*request = gdam_request_set_args(
		source_id,
		num_args,
		args,
		handle_set_command_success,
		handle_set_command_failure,
		NULL
	);
	gdam_arg_free_array(num_args, args);
	g_free(id);
	return *request ? TRUE : FALSE;
}

/*
 *            _     _
 *   __ _  __| | __| |
 *  / _` |/ _` |/ _` |
 * | (_| | (_| | (_| |
 *  \__,_|\__,_|\__,_|
 * Add a source to a mixer.
 */

static void handle_add_success(gpointer unused)
{
}
static void handle_add_failure(gpointer unused)
{
}
static gboolean handle_add_command(LineTokenizer* tokenizer,
                                   CliInfo*       info,
				   GdamRequest**  request)
{
	guint mix_id = info->mixer_id;
	guint sub_id;
	char* num;
	num = line_tokenizer_get_word(tokenizer);
	if (num == NULL) {
		fprintf(stderr, "no source id found...\n");
		return FALSE;
	}
	sub_id = (guint)strtol(num, NULL, 10);
	g_free(num);
	num = line_tokenizer_get_word(tokenizer);
	if (num != NULL) {
		fprintf(stderr, "got word mix id word `%s'", num);
		mix_id = (guint)strtol(num, NULL, 10);
		g_free(num);
	}
	gdam_debug("request_add_mix: mix=%d, sub=%d", mix_id, sub_id);
	*request = gdam_request_add_mix(
				mix_id,
				sub_id,
				(GdamTime) -1,
				handle_add_success,
				handle_add_failure,
				NULL);
	return *request ? TRUE : FALSE;
}

/*
 *      _ _
 *  ___(_) | ___ _ __   ___ ___
 * / __| | |/ _ \ '_ \ / __/ _ \
 * \__ \ | |  __/ | | | (_|  __/
 * |___/_|_|\___|_| |_|\___\___|
 * 
 */

static void handle_silence_success(gpointer unused)
{
	gdam_debug("silenced mixer successfully");
}
static void handle_silence_failure(gpointer unused)
{
	g_message("Failed to silence the mixer");
}
static gboolean handle_silence_command(LineTokenizer* tokenizer,
                                   CliInfo*       info,
				   GdamRequest**  request)
{
	guint mix_id = info->mixer_id;
	char* num;
	GdamArg arg[1];
	num = line_tokenizer_get_word(tokenizer);
	if (num != NULL) {
		fprintf(stderr, "got word mix id word `%s'", num);
		mix_id = (guint)strtol(num, NULL, 10);
		g_free(num);
	}
	gdam_debug("request_silence_mix: mix=%d", mix_id);
	arg[0].name = "mixed";
	arg[0].type = GDAM_ARG_TYPE_INT64_ARRAY;
	GDAM_ARG_INT64_ARRAY(arg).length = 0;
	GDAM_ARG_INT64_ARRAY(arg).array = NULL;
	*request = gdam_request_set_args(
				mix_id,
				1,
				arg,
				handle_silence_success,
				handle_silence_failure,
				NULL);
	return *request ? TRUE : FALSE;
}

/*
 *            _
 *  _ __ ___ (_)_  _____ _ __
 * | '_ ` _ \| \ \/ / _ \ '__|
 * | | | | | | |>  <  __/ |
 * |_| |_| |_|_/_/\_\___|_|
 * Print the mixer id.
 */
static gboolean handle_mixer_command(LineTokenizer* tokenizer,
                                     CliInfo*       info,
				     GdamRequest**  request)
{
	fprintf(stderr, "mixer id is %d.\n", info->mixer_id);
	return TRUE;
}

static gboolean handle_device_command(LineTokenizer* tokenizer,
                                      CliInfo*       info,
				      GdamRequest**  request)
{
	fprintf(stderr, "device id is %d.\n", info->device_id);
	return TRUE;
}
/*
 *      _      _             _
 *   __| | ___| |_ __ _  ___| |__
 *  / _` |/ _ \ __/ _` |/ __| '_ \
 * | (_| |  __/ || (_| | (__| | | |
 *  \__,_|\___|\__\__,_|\___|_| |_|
 * Remove a source from its parent.
 */
static void handle_detach_success(gpointer unused)
{
}
static void handle_detach_failure(gpointer unused)
{
}
static gboolean handle_detach_command(LineTokenizer* tokenizer,
                                      CliInfo*       info,
				      GdamRequest**  request)
{
	guint source_id;
	char* num;
	num = line_tokenizer_get_word(tokenizer);
	if (num == NULL) {
		fprintf(stderr, "what do you want to detach?\n");
		return FALSE;
	}
	source_id = (guint) strtoul(num, NULL, 10);
	g_free(num);

	*request = gdam_request_detach_source(source_id,
	                                      handle_detach_success,
					      handle_detach_failure,
					      NULL);
	if (!*request) {
		fprintf(stderr, "unable to detach source %d.\n",
			source_id);
		return FALSE;
	}
	return TRUE;
}

static gboolean handle_feed_command(LineTokenizer* tokenizer,
                                    CliInfo*       info,
				    GdamRequest**  request)
{
	guint device_id, mixer_id;
	char* num;
	num = line_tokenizer_get_word(tokenizer);
	if (num == NULL) {
		device_id = info->device_id;
		mixer_id = info->mixer_id;
		g_message("reopening default device %d", device_id);
	} else {
		device_id = atoi(num);
		g_free(num);

		num = line_tokenizer_get_word(tokenizer);
		if (num == NULL) {
			mixer_id = info->mixer_id;
			g_message("reopening default device %d", device_id);
		} else {
			mixer_id = atoi(num);
			g_free(num);
		}
	}

	*request = gdam_request_feed_sink(device_id,
			                  mixer_id,
	                                      NULL,
					      NULL,
					      NULL);
	if (!*request) {
		g_warning("unable to feed sink %d", device_id);
		return FALSE;
	}
	return TRUE;
}

static gboolean handle_reopen_command(LineTokenizer* tokenizer,
                                      CliInfo*       info,
				      GdamRequest**  request)
{
	guint device_id;
	char* num;
	num = line_tokenizer_get_word(tokenizer);
	if (num == NULL) {
		device_id = info->device_id;
		g_message("reopening default device %d", device_id);
	} else {
		device_id = atoi(num);
		g_free(num);
	}

	*request = gdam_request_reopen_sink(device_id,
	                                      NULL,
					      NULL,
					      NULL);
	if (!*request) {
		g_warning("unable to detach source %d", device_id);
		return FALSE;
	}
	return TRUE;
}


/*
 *  _
 * | |___
 * | / __|
 * | \__ \
 * |_|___/
 * 
 */
static void
process_ls_success(int num_args, GdamArg* args, gpointer user_data)
{
	int i;
	fprintf(stderr, "source %d.\n", GPOINTER_TO_INT(user_data));
	for (i = 0; i < num_args; i++) {
		char* arg_value;
		arg_value = gdam_arg_val_as_string(args + i);
		fprintf(stderr, "  arg %d: %s => %s\n",
		                i,
				args[i].name,
				arg_value);
		g_free(arg_value);
	}
}
static void process_ls_failure(gpointer user_data)
{
	fprintf(stderr, "ls on source %d failed.\n",
			GPOINTER_TO_INT(user_data));
}
static gboolean handle_ls_command(LineTokenizer* tokenizer,
                                  CliInfo*       info,
				  GdamRequest**  request)
{
	guint source_id;
	char* num = NULL;
	gboolean only_vital = TRUE;
	while (num == NULL) {
		char* arg = line_tokenizer_get_word(tokenizer);
		if (arg == NULL) {
			fprintf(stderr, "no source id found...\n");
			return FALSE;
		}
		if (arg[0] == '-') {
			switch (arg[1]) {
				case 'a': only_vital = FALSE; break;
			}
			g_free(arg);
			continue;
		}
		num = arg;
	}
	source_id = (int) strtol(num, NULL, 10);
	*request = gdam_request_get_all_args(source_id,
	                                     only_vital,
	                                     process_ls_success,
					     process_ls_failure,
					     GINT_TO_POINTER(source_id));
	return *request ? TRUE : FALSE;
}

/*
 *            _ _
 *   ___ __ _| | |
 *  / __/ _` | | |
 * | (_| (_| | | |
 *  \___\__,_|_|_|
 * 
 */

static void
process_call_success(int num_args, GdamArg* args, gpointer user_data)
{
	int i;
	fprintf(stderr, "call: source %d.\n", GPOINTER_TO_INT(user_data));
	for (i = 0; i < num_args; i++) {
		char* arg_value;
		arg_value = gdam_arg_val_as_string(args + i);
		fprintf(stderr, "  arg %d: %s => %s\n",
		                i,
				args[i].name,
				arg_value);
		g_free(arg_value);
	}
}
static void process_call_failure(gpointer unused)
{
	gdam_debug("call command failed");
}

static gboolean handle_call_command(LineTokenizer* tokenizer,
                                    CliInfo*       info,
				    GdamRequest**  request)
{
	char* num = line_tokenizer_get_word(tokenizer);
	char* func_name = line_tokenizer_get_word(tokenizer);
	int num_args = 0, max_args = 16;
	int source_id;
	GdamArg* args = g_new(GdamArg, max_args);
	while (1) {
		if (num_args == max_args) {
			max_args *= 2;
			args = g_renew(GdamArg, args, max_args);
		}
		if (!line_tokenizer_tokenize(tokenizer,
					     args + num_args)) {
			break;
		}
		num_args++;
	}
	g_return_val_if_fail(func_name != NULL, FALSE);
	source_id = (int) strtol(num, NULL, 10);
	*request = gdam_request_call_function(GDAM_PROTOCOL_SOURCE_ARG,
	                                      source_id,
					      func_name,
					      num_args,
					      args,
	                                      process_call_success,
					      process_call_failure,
					      GINT_TO_POINTER(source_id));
	return *request ? TRUE : FALSE;
}

/*
 *      _
 *  ___| | ___  ___ _ __
 * / __| |/ _ \/ _ \ '_ \
 * \__ \ |  __/  __/ |_) |
 * |___/_|\___|\___| .__/
 *                 |_|
 * sleep for a number of seconds (possibly fractional).
 */
static gboolean handle_sleep_command(LineTokenizer* tokenizer,
                                     CliInfo*       info,
				     GdamRequest**  request)
{
	char* num_str;
	char* endp;
	double secs;
	num_str = line_tokenizer_get_word(tokenizer);
	if (num_str == NULL) {
		g_warning("sleep needs an argument");
		return FALSE;
	}
	secs = strtod(num_str, &endp);
	if (endp == num_str || *endp != '\0') {
		g_warning("error parsing number");
		return FALSE;
	}
	usleep(secs * 1000000);
	return TRUE;
}
/*
 *            _ _
 *   _____  _(_) |_
 *  / _ \ \/ / | __|
 * |  __/>  <| | |_
 *  \___/_/\_\_|\__|
 * Exit (or quit) command.
 */
static gboolean handle_exit_command(LineTokenizer* tokenizer,
                                    CliInfo*       info,
				    GdamRequest**  request)
{
	int exit_code = 0;
	char* num;
	num = line_tokenizer_get_word(tokenizer);
	if (num != NULL)
		exit_code = strtol(num, NULL, 10);
	exit(exit_code);
}

/*
 *  _          _
 * | |__   ___| |_ __
 * | '_ \ / _ \ | '_ \
 * | | | |  __/ | |_) |
 * |_| |_|\___|_| .__/
 *              |_|
 * Help command.
 */
static gboolean handle_help_command(LineTokenizer* tokenizer,
                                    CliInfo*       info,
				    GdamRequest**  request)
{
	char* topic = line_tokenizer_get_word(tokenizer);
	if (topic) {
		cli_help(topic);
		g_free(topic);
	} else {
		fprintf(stderr,
			"Commands:\n\n"
			"\tls      - list a source's information.\n"
			"\tmixer   - get the default mixer id.\n"
			"\tdevice  - get the default device id.\n"
			"\tsource  - make a new sound source.\n"
			"\tcall    - make a server-side function call.\n"
			"\tset     - set a sound source's properties.\n"
			"\treopen  - close/open a sound device.\n"
			"\tdetach  - parent a source from its parent.\n"
			"\tadd     - add a source to a mix.\n"
			"\tsilence - remove all sources from a mixer.\n"
			"\thelp    - online help.\n"
			"\tsleep # - sleep a number of seconds.\n"
			"\n"
			"Run `help COMMAND' for more help on one of "
			"the above.\n"
		);
	}
	return TRUE;
}

/*=====================================================================*/
static void register_command_handler(const char*    command,
                                     CommandHandler handler)
{
	g_assert(g_hash_table_lookup(command_handlers, command) == NULL);
	g_hash_table_insert(command_handlers,
	                    g_strdup(command),
			    (gpointer) handler);
}
static CommandHandler lookup_command_handler(const char* command)
{
	if (!command)
		return NULL;
	return (CommandHandler) g_hash_table_lookup(command_handlers, command);
}

static gboolean make_request_from_command(const char*    command,
                                          LineTokenizer* tokenizer,
                                          CliInfo*       info,
				          GdamRequest**  request)
{
	CommandHandler handler = lookup_command_handler(command);
	if (!handler) {
		fprintf(stderr, "unknown command type %s.\n", command);
		return FALSE;
	}
	*request = NULL;
	return (*handler)(tokenizer, info, request);
}

/* Build default commands. */
static void initialize_default_comamnds()
{
	register_command_handler("source", handle_source_command);
	register_command_handler("detach", handle_detach_command);
	register_command_handler("feed", handle_feed_command);
	register_command_handler("add", handle_add_command);
	register_command_handler("set", handle_set_command);
	register_command_handler("ls", handle_ls_command);
	register_command_handler("call", handle_call_command);
	register_command_handler("silence", handle_silence_command);
	register_command_handler("mixer", handle_mixer_command);
	register_command_handler("device", handle_device_command);
	register_command_handler("reopen", handle_reopen_command);
	register_command_handler("quit", handle_exit_command);
	register_command_handler("exit", handle_exit_command);
	register_command_handler("help", handle_help_command);
	register_command_handler("sleep", handle_sleep_command);
}
	
static void usage(const char* prog_name)
{
	fprintf(stderr, "usage: %s [gdam-options]\n\n"
	                "Client-Line Interface to gdam.\n\n"
			"Gdam-Options:\n"
			"%s",
			prog_name,
			gdam_channel_usage_message);
	exit(1);
}

static gboolean is_blank(const char* str)
{
	while (*str && isspace(*str))
		str++;
	return *str == 0 ? TRUE : FALSE;
}

static char* dollar_expand(const char* orig, GHashTable* vars)
{
	GString* string;
	const char* at;
	char* rv;
	string = g_string_new("");
	for (at = orig; *at != '\0'; at++) {
		if (*at == '$') {
			const char* end_name = at+1;
			char* name;
			char* value;
			while (*end_name && isalnum(*end_name))
				end_name++;
			if (end_name == at + 1) {
				g_string_append_c(string, '$');
				continue;
			}
			name = (char*) alloca(end_name + 1 - at);
			memcpy(name, at + 1, end_name - (at + 1));
			name[end_name - (at + 1)] = '\0';
			value = g_hash_table_lookup(vars, name);
			if (value == NULL) {
				g_warning("no variable %s", name);
				g_string_append_c(string, '$');
				continue;
			}
			g_string_append(string, value);
			at = end_name - 1;
		} else {
			g_string_append_c(string, *at);
		}
	}
	rv = string->str;
	g_string_free(string, FALSE);
	return rv;
}

void process_file(CliInfo* cli_info, const char* fname);

void process_line(CliInfo* cli_info, const char* line)
{
	LineTokenizer* tokenizer;
	char* command;
	char* expanded_line;
	GdamRequest* request;
	//tokenizer->variables = variables;
	expanded_line = dollar_expand(line, variables);
	tokenizer = line_tokenizer_new(expanded_line);

	command = line_tokenizer_get_word(tokenizer);
	if (command) {
		if (strcmp(command, ".") == 0) {
			char* fname = line_tokenizer_get_word(tokenizer);
			g_return_if_fail(fname != NULL);
		 	process_file(cli_info, fname);
			g_free(expanded_line);
			return;
		}
		if (!make_request_from_command(command,
					       tokenizer,
					       cli_info,
					       &request)) {
			g_warning("error parsing %s", line);
			g_free(command);
			g_free(expanded_line);
			line_tokenizer_destroy(tokenizer);
			return;
		}
		if (request) {
			if (!gdam_channel_process_request(cli_info->channel,
							  request))
			g_error("error processing request");
		}
		g_free(command);
	}
	line_tokenizer_destroy(tokenizer);
	g_free(expanded_line);
}
void process_file(CliInfo* cli_info, const char* fname)
{
	char buf[4096*8];
	FILE* fp;
	fp = fopen(fname, "r");
	if (!fp) {
		fprintf(stderr, "error opening %s\n", fname);
		return;
	}
	while (fgets(buf,sizeof(buf),fp))
		process_line(cli_info, buf);
	fclose(fp);
}

static void
open_device(GdamChannel*   channel,
            const char*    device_name,
	    int*           device_id,
	    int*           mixer_id)
{
	GdamRequestGetIntStruct int_val;
	g_return_if_fail(device_name != NULL);
	gdam_channel_process_request(
		channel,
		gdam_request_get_server_side_int
			  ("devices",
			   device_name,
			   gdam_request_get_int,
			   gdam_request_get_failure,
			   &int_val)
	);
	if (!int_val.was_success) {
		gdam_channel_process_request(
			channel,
			gdam_request_create_sink
				(device_name,
				 gdam_request_get_int,
				 gdam_request_get_failure,
				 &int_val)
		);
		if (!int_val.was_success)
			g_error("error opening device %s", device_name);
	}
	g_assert(int_val.was_success);
	*device_id = int_val.rv;


	gdam_channel_process_request(
		channel,
		gdam_request_get_server_side_int
			  ("device-feeds",
			   device_name,
			   gdam_request_get_int,
			   gdam_request_get_failure,
			   &int_val)
	);
	if (!int_val.was_success) {
		GdamRequest** package = g_new(GdamRequest*, 3);
		package[0] = gdam_request_create_source(
					"GdamMix",
					gdam_request_get_int,
					gdam_request_get_failure,
					&int_val,
					1,
					"mode",
					GDAM_ARG_TYPE_STRING,
					"eternal"
				);
		package[1] = gdam_request_create_source(
					"GdamRecorderExternal",
					NULL, NULL, NULL,
					1,
					"filtered",
					GDAM_ARG_TYPE_INT,
					-1);
		package[2] = gdam_request_feed_sink(
					*device_id,
					-2,
					NULL,
					NULL,
					"eternal"
				);
		gdam_channel_process_request(
			channel,
			gdam_request_package_v
				(3,
				 package,
				 NULL,
				 NULL,
				 NULL)
		);
		if (!int_val.was_success)
			g_error("error opening device %s", device_name);
	}
	g_assert(int_val.was_success);
	*mixer_id = int_val.rv;
}


int main(int argc, char** argv)
{
	GdamChannel* channel;
	int device_id = 0, mixer_id = 0;
	char* line;
	char* prompt = "gdam-cli> ";
	int i;
	gboolean no_devices = FALSE, device_opened = FALSE;
	GSList* expressions = NULL;
	CliInfo cli_info;
	if (argc >= 2) {
		if (strcmp(argv[1], "--help") == 0
		 || strcmp(argv[1], "-help") == 0
		 || strcmp(argv[1], "-h") == 0)
		 	usage(argv[0]);
		if (strcmp(argv[1], "-v") == 0
		 || strcmp(argv[1], "--version") == 0) {
		 	printf(
			     "gdamcli version %s.\n"
		             "(C) Copyright Dave Benson, Geoff Matters. 2000.\n"
			     "\n"
			     "THIS PROGRAM COMES WITH NO WARRANTY! "
			     "USE AT YOUR OWN RISK.\n",
			     VERSION);
			exit(0);
		}
	}
	initial_empty_hashtables();
	if (!gdam_channel_init(&argc, &argv, &channel)) {
		g_error("error creating connection to client");
		return 1;
	}
	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "--device") == 0
		 || strcmp(argv[i], "-device") == 0) {
		 	open_device(channel, argv[++i], &device_id, &mixer_id);
			device_opened = TRUE;
			continue;
		}
		if (strcmp(argv[i],"-e") == 0) {
			if (argv[++i] == NULL) {
				g_warning("no command after `-e'");
				return 1;
			}
			expressions = g_slist_prepend(expressions, argv[i]);
			continue;
		}
		if (argv[i][0] == '-') {
			g_warning("unrecognized option: %s", argv[i]);
			continue;
		}
		expressions = g_slist_prepend(expressions,
		                              g_strdup_printf(". %s", argv[i]));
	}
	if (!device_opened && !no_devices) {
		char* device = getenv("GDAM_DEVICE");
		if (device) {
		 	open_device(channel, device, &device_id, &mixer_id);
			device_opened = TRUE;
		}
	}
	if (!device_opened && !no_devices) {
		g_message("Using /dev/dsp since no --device was given "
		          "but --no-device was not specified");
		open_device(channel, "/dev/dsp", &device_id, &mixer_id);
		device_opened = TRUE;
	}
	cli_info.interactive = 1;
	cli_info.device_id = device_id;
	cli_info.mixer_id = mixer_id;
	cli_info.channel = channel;
	g_hash_table_insert(variables,
	                    "mixer",
			    g_strdup_printf("%d", mixer_id));
	g_hash_table_insert(variables,
	                    "device",
			    g_strdup_printf("%d", device_id));


	initialize_default_comamnds();

	if (expressions) {
		GSList* list;
		for (list = expressions; list; list = list->next) {
			char* line = list->data;
			process_line(&cli_info, line);
		}
		return 0;
	}


	while ((line = readline(prompt)) != NULL) {
		if (!is_blank(line))
			add_history(line);
		process_line(&cli_info, line);
	}

	return 0;
}
