/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamchannel.h"
#include "gdamconnect.h"
#include <stdlib.h>
#include <gtk/gtkmain.h>
#include <gdam/gdam-protocol.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <sys/poll.h>


const char* default_addr_spec = "localhost:6003";
GdamBuffer default_auth_info;

GdamChannel* gdam_channel_new()
{
	if (default_addr_spec == NULL) {
		g_warning("gdam_channel_new called before gdam_init()");
		return NULL;
	}
	return gdam_channel_new_full(default_addr_spec, &default_auth_info);
}

static gboolean default_auth_handler(GdamChannel* channel, GdamBuffer* resp)
{
	gdam_debug("default_auth_handler:  proceeding");
	return TRUE;
}
GdamChannel* gdam_channel_new_full(const char* addr_spec,
                                   GdamBuffer* auth_info)
{
	GdamChannel* channel;
	int fd;
	int is_connecting = 0;

	switch (gdam_connect(addr_spec, &fd)) {
		case GDAM_CONNECT_CONNECTING:
			is_connecting = 1;
			break;
		case GDAM_CONNECT_CONNECTED:
			is_connecting = 0;
			break;
		case GDAM_CONNECT_FAILED:
			g_warning("error connecting to %s", addr_spec);
			return NULL;
	}
			
	channel = g_new(GdamChannel, 1);
	gdam_buffer_construct(&channel->outgoing);
	gdam_buffer_construct(&channel->incoming);
	channel->first_request = channel->last_request = NULL;
	channel->first_pending = channel->last_pending = NULL;
	channel->num_pending = 0;
	channel->max_pending = 5;
	channel->poll_fd.fd = fd;
	channel->source_id = 0;
	channel->ref_count = 1;
	channel->is_connecting = is_connecting;
	channel->quit_main = 0;
	channel->auth_handler = default_auth_handler;
	channel->event_to_info_map = g_hash_table_new(NULL, NULL);
	{
		int length;
		g_assert(gdam_buffer_peek_int(auth_info, &length));
		g_assert(auth_info->size == length);
	}
	gdam_buffer_append_buffer(&channel->outgoing, auth_info, FALSE);

	/* TODO: register auth handshake (as pending request) */
	return channel;
}


static gboolean gdam_channel_prepare(gpointer  source_data, 
			             GTimeVal *current_time,
			             gint     *timeout,
			             gpointer  user_data)
{
	GdamChannel* channel = (GdamChannel*)user_data;
	if (channel->is_connecting) {
		channel->poll_fd.events = G_IO_IN|G_IO_OUT;
	} else if (channel->outgoing.size > 0) {
		channel->poll_fd.events = G_IO_IN|G_IO_OUT;
	} else {
		channel->poll_fd.events = G_IO_IN;
	}

	*timeout = -1;
	channel->poll_fd.revents = 0;
	return FALSE;
}
static gboolean gdam_channel_check(gpointer  source_data, 
			           GTimeVal *current_time,
			           gpointer  user_data)
{
	GdamChannel* channel = (GdamChannel*)user_data;
	if (channel->poll_fd.revents)
		return TRUE;
	return FALSE;
}
static gboolean is_event_type(int type)
{	
	if (type > GDAM_EVENT_FIRST
	 && type < GDAM_MESSAGE_LAST_MESSAGE)
	 	return TRUE;
	return FALSE;
}

void gdam_channel_register_handler(GdamChannel* channel, 
                                  GdamChannelEventHandler* handler)
{
	g_hash_table_insert(channel->event_to_info_map,
	                    GUINT_TO_POINTER(handler->connection_id),
			    handler);
}
void gdam_channel_unregister_handler(GdamChannel* channel,
                                     guint connection_id)
{
	g_hash_table_remove(channel->event_to_info_map,
	                    GUINT_TO_POINTER(connection_id));
}

static gboolean one_shot_simple_marshal(GdamChannelEventHandler* handler,
	                                GdamBuffer* buffer,
					guint connection_id)
{
	handler->callback(handler->user_data);
	return TRUE;
}
static gboolean repeating_simple_marshal(GdamChannelEventHandler* handler,
	                                 GdamBuffer* buffer,
					 guint connection_id)
{
	handler->callback(handler->user_data);
	return FALSE;
}

GdamChannelEventHandler*
gdam_channel_simple_handler_new(guint connection_id,
                                gboolean one_shot,
                                GHookFunc func,
				gpointer func_data)
{
	GdamChannelEventHandler* handler;
	handler = g_new(GdamChannelEventHandler, 1);
	handler->connection_id = connection_id;
	handler->callback = func;
	handler->user_data = func_data;
	if (one_shot)
		handler->marshal = one_shot_simple_marshal;
	else
		handler->marshal = repeating_simple_marshal;
	return handler;
}

static gboolean one_shot_time_marshal(GdamChannelEventHandler* handler,
	                                GdamBuffer* buffer,
					guint connection_id)
{
	GdamRequestGotTimeFunc func = (GdamRequestGotTimeFunc)
						(handler->callback);
	GdamTime time;
	if (!gdam_buffer_parse_time(buffer, &time))
		return TRUE;
	(*func)(time, handler->user_data);
	return TRUE;
}
static gboolean repeating_time_marshal(GdamChannelEventHandler* handler,
	                                 GdamBuffer* buffer,
					 guint connection_id)
{
	GdamRequestGotTimeFunc func = (GdamRequestGotTimeFunc)
						(handler->callback);
	GdamTime time;
	if (!gdam_buffer_parse_time(buffer, &time)) {
		g_warning("error parsing time in periodic timer");
		return TRUE;		/* causes it to stop happening... */
	}
	(*func)(time, handler->user_data);
	return FALSE;
}

GdamChannelEventHandler* gdam_channel_time_handler_new
                                                  (guint        connection_id,
						   gboolean     one_shot,
	                                           GdamRequestGotTimeFunc func,
						   gpointer     func_data)
{
	GdamChannelEventHandler* handler;
	handler = g_new(GdamChannelEventHandler, 1);
	handler->connection_id = connection_id;
	handler->callback = (GHookFunc)func;
	handler->user_data = func_data;
	if (one_shot)
		handler->marshal = one_shot_time_marshal;
	else
		handler->marshal = repeating_time_marshal;
	return handler;
}


/* INT */
static gboolean one_shot_int_marshal(GdamChannelEventHandler* handler,
	                                GdamBuffer* buffer,
					guint connection_id)
{
	GdamRequestGotIntFunc func = (GdamRequestGotIntFunc)
						(handler->callback);
	int i;
	if (!gdam_buffer_parse_int(buffer, &i))
		return TRUE;
	(*func)(i, handler->user_data);
	return TRUE;
}
static gboolean repeating_int_marshal(GdamChannelEventHandler* handler,
	                                 GdamBuffer* buffer,
					 guint connection_id)
{
	GdamRequestGotIntFunc func = (GdamRequestGotIntFunc)
						(handler->callback);
	int i;
	if (!gdam_buffer_parse_int(buffer, &i)) {
		g_warning("error parsing int in periodic timer");
		return TRUE;		/* causes it to stop happening... */
	}
	(*func)(i, handler->user_data);
	return FALSE;
}

GdamChannelEventHandler* gdam_channel_int_handler_new
                                                  (guint        connection_id,
						   gboolean     one_shot,
	                                           GdamRequestGotIntFunc func,
						   gpointer     func_data)
{
	GdamChannelEventHandler* handler;
	handler = g_new(GdamChannelEventHandler, 1);
	handler->connection_id = connection_id;
	handler->callback = (GHookFunc)func;
	handler->user_data = func_data;
	if (one_shot)
		handler->marshal = one_shot_int_marshal;
	else
		handler->marshal = repeating_int_marshal;
	return handler;
}
static gboolean gdam_channel_dispatch_event(GdamChannel* channel,
                                            GdamBuffer*  command)
{
	/* Lookup handler. */
	GdamChannelEventHandler* info;
	int type;
	guint id;
	GHashTable* event_to_info_map = channel->event_to_info_map;

	if (!gdam_buffer_parse_int(command, &type)
	 || !gdam_buffer_parse_int(command, &id)) {
	 	gdam_debug("dispatch_event: too short");
		return FALSE;
	}

	info = (GdamChannelEventHandler*)g_hash_table_lookup(
						event_to_info_map,
	                                        GUINT_TO_POINTER(id)
					);

	if (!info) {
		gdam_debug("event %d of type %d occurred when not expected",
			   id, type);
		return FALSE;
	}
	/* Invoke handler. */
	if (info->marshal(info, command, id)) {
		g_free(info);
		g_hash_table_remove(event_to_info_map, GUINT_TO_POINTER(id));
	}
	return TRUE;
}

static gboolean gdam_channel_dispatch_response(GdamChannel* channel,
                                               GdamBuffer*  command)
{
	GdamRequest* pending = (GdamRequest*)channel->first_pending->data;
	channel->first_pending = g_slist_remove(channel->first_pending,
	                                        pending);
	channel->num_pending--;
	if (!channel->first_pending)
		channel->last_pending = NULL;
	if (!pending->marshal(pending, command))
		return FALSE;
	gdam_request_destroy(pending);
	return TRUE;
}

/* If there is enough data in the buffer to dispatch,
 * we do so... */
static gboolean gdam_channel_maybe_dispatch_input(GdamChannel* channel)
{
	GdamBuffer* incoming = &channel->incoming;
	while (1) {
		GdamBuffer cmd;
		int length, type;
		if (!gdam_buffer_peek_int(incoming, &length))
			break;
		if (incoming->size < length)
			break;
		gdam_buffer_construct(&cmd);
		gdam_buffer_send_buffer(&cmd, incoming, length, TRUE);
		g_assert(length == cmd.size);
		gdam_buffer_parse_int(&cmd, &length);
		if (channel->auth_handler) {
			if (!channel->auth_handler(channel, &cmd)) {
				gdam_buffer_destruct(&cmd);
				return FALSE;
			}
			channel->auth_handler = NULL;
			gdam_buffer_destruct(&cmd);
			continue;
		}
		gdam_buffer_peek_int(&cmd, &type);
		if (is_event_type(type)) {
			if (!gdam_channel_dispatch_event(channel, &cmd)) {
				gdam_buffer_destruct(&cmd);
				return FALSE;
			}
		} else {
			if (!gdam_channel_dispatch_response(channel, &cmd)) {
				gdam_buffer_destruct(&cmd);
				return FALSE;
			}
		}
		gdam_buffer_destruct(&cmd);
	}
	return TRUE;
}

static void maybe_write_to_outgoing(GdamChannel* channel)
{
	while (channel->num_pending < channel->max_pending
	    && channel->first_request != NULL) {
		GSList* old_first = channel->first_request;
		GdamRequest* request = (GdamRequest*)(old_first->data);
		GdamBuffer* req_buf = &request->request_command;
		int length;

		g_assert(gdam_buffer_peek_int(req_buf, &length));
		g_assert(req_buf->size == length);

		gdam_buffer_append_buffer(&channel->outgoing, req_buf, TRUE);

		/* Now, horrible first manipulation to remove
		 * `first_request' and append it as `last_pending'...
		 * (This looks horrible but it is efficient in that
		 *  it doesn't create or destroy a list node. */
		channel->first_request = old_first->next;
		if (!channel->first_request)
			channel->last_request = NULL;
		old_first->next = NULL;
		if (channel->last_pending)
			channel->last_pending->next = old_first;
		else
			channel->first_pending = old_first;
		channel->last_pending = old_first;
		channel->num_pending++;
	}
}

static gboolean gdam_channel_dispatch(gpointer  source_data, 
			              GTimeVal *current_time,
			              gpointer  user_data)
{
	GdamChannel* channel = (GdamChannel*)user_data;
	int fd = channel->poll_fd.fd;
	if (channel->is_connecting) {
		if (!gdam_connect_finish(fd)) {
			if (channel->quit_main)
				exit(0);
			return FALSE;
		}
		channel->is_connecting = 0;
		return TRUE;
	}
	/*
	gdam_debug("dispatch io-event: fd=%d; ev=%d [in=%d, out=%d]", 
		channel->poll_fd.fd, channel->poll_fd.revents,
		G_IO_IN, G_IO_OUT);
	*/
	if ((channel->poll_fd.revents & G_IO_IN) == G_IO_IN) {
		if (!gdam_buffer_read_one(&channel->incoming, fd))
			return FALSE;
		if (!gdam_channel_maybe_dispatch_input(channel))
			return FALSE;
	}
	if ((channel->poll_fd.revents & G_IO_OUT) == G_IO_OUT) {
		if (!gdam_buffer_write_one(&channel->outgoing, fd))
			return FALSE;
		maybe_write_to_outgoing(channel);
	}
	return TRUE;
}


static GSourceFuncs channel_source_funcs = {
	gdam_channel_prepare,
	gdam_channel_check,
	gdam_channel_dispatch,
	NULL
};


static void gdam_channel_detach_notify(gpointer chan)
{
	GdamChannel* channel = (GdamChannel*)chan;
	if (channel->source_id == 0) return;
	channel->source_id = 0;
	g_main_remove_poll(&channel->poll_fd);
	gdam_channel_unref(channel);
}

void gdam_channel_ref(GdamChannel* channel)
{
	++channel->ref_count;
}

void gdam_channel_unref(GdamChannel* channel)
{
	if (--channel->ref_count == 0)
		gdam_channel_destroy(channel);
}

/* Attach to the main loop... you have to do this for
 * events to get processed. */
void gdam_channel_attach(GdamChannel* channel)
{
	g_return_if_fail(channel->source_id == 0);
	channel->source_id = g_source_add(G_PRIORITY_DEFAULT,
	                                  TRUE,
					  &channel_source_funcs,
					  NULL,
					  channel,
					  gdam_channel_detach_notify);
	g_main_add_poll(&channel->poll_fd, G_PRIORITY_DEFAULT);
	++channel->ref_count;
}

void gdam_channel_detach(GdamChannel* channel)
{
	int source_id = channel->source_id;
	if (source_id == 0)
		return;
	channel->source_id = 0;
	g_source_remove(source_id);
	g_main_remove_poll(&channel->poll_fd);
}


/* Delete us (safe whether or not attached). */
void gdam_channel_destroy(GdamChannel* channel)
{
	gdam_channel_detach(channel);
	gdam_buffer_destruct(&channel->outgoing);
	gdam_buffer_destruct(&channel->incoming);
	g_slist_foreach(channel->first_request,
	                (GFunc)gdam_request_destroy,
			NULL);
	g_slist_foreach(channel->first_pending,
	                (GFunc)gdam_request_destroy,
			NULL);
	g_slist_free(channel->first_request);
	g_slist_free(channel->first_pending);
	if (channel->quit_main)
		exit(0);
	g_free(channel);
}


gboolean gdam_channel_append_request   (GdamChannel* channel,
                                        GdamRequest* request)
{
	GSList* new_last = g_slist_prepend(NULL, request);
	if (channel->last_request)
		channel->last_request->next = new_last;
	else
		channel->first_request = new_last;
	channel->last_request = new_last;
	maybe_write_to_outgoing(channel);
	return TRUE;
}


gboolean gdam_channel_prepend_request   (GdamChannel* channel,
                                        GdamRequest* request)
{
	GSList* new_first = g_slist_prepend(channel->first_request, request);
	if (!channel->last_request)
		channel->last_request = new_first;
	channel->first_request = new_first;
	maybe_write_to_outgoing(channel);
	return TRUE;
}

static gboolean flush_pending(GdamChannel* channel)
{
	struct pollfd fds;
	fds.fd = channel->poll_fd.fd;
	while (channel->first_pending) {
		char buf[4096];
		int num_read;
		fds.events = POLLIN;
		if (channel->outgoing.size > 0)
			fds.events |= POLLOUT;
		if (poll(&fds, 1, -1) < 0) {
			if (errno == EINTR)
				continue;
			perror("polling in flush_pending");
			return FALSE;
		}
		if (fds.revents & POLLIN) {
			num_read = read(channel->poll_fd.fd, buf, sizeof(buf));
			if (num_read == 0) {
				g_message("flush_pending_requests: "
						"eof from server");
				return FALSE;
			}
			if (num_read < 0) {
				if (errno == EINTR)
					continue;
				if (errno == EWOULDBLOCK) {
					g_warning(
					   "non-blocking socket going into "
					   "busy loop (flush-pending)");
					continue;
				}
				perror("reading doing synchronous command");
				return FALSE;
			}
			gdam_buffer_append(&channel->incoming, buf, num_read);
			if (!gdam_channel_maybe_dispatch_input(channel))
				return FALSE;
		}
		if (fds.revents & POLLOUT)
			if (!gdam_buffer_write_one(&channel->outgoing, fds.fd))
				return FALSE;
	}
	return TRUE;
}
	
static void poll_for_readwrite(int fd)
{
	struct pollfd fds;
	fds.fd = fd;
	fds.events = POLLIN|POLLOUT;
	poll(&fds, 1, -1);
}

gboolean gdam_channel_process_request(GdamChannel* channel, 
                                      GdamRequest* request)
{
	static gboolean processing_request = FALSE;

	gboolean rv;

	if (processing_request) {
		g_error("process_request was reinvoked from a request "
			"terminate handler");
		return FALSE;
	}

	processing_request = TRUE;
	if (channel->is_connecting) {
		gdam_debug("[process_request] connecting...");
		poll_for_readwrite(channel->poll_fd.fd);
		if (!gdam_connect_finish(channel->poll_fd.fd)) {
			if (channel->quit_main)
				exit(0);
			return FALSE;
		}
		channel->is_connecting = 0;
	}
	if (channel->auth_handler)
		gdam_channel_append_request(channel, request);
	else
		gdam_channel_prepend_request(channel, request);
	rv = flush_pending(channel);
	processing_request = FALSE;
	return rv;
}

