/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamsoundspec.h"
#include <gdam/gdam-protocol.h>
#include <gdam/gdamsample.h>

/*
 *  __  __ _____  _______ ____
 * |  \/  |_ _\ \/ / ____|  _ \
 * | |\/| || | \  /|  _| | |_) |
 * | |  | || | /  \| |___|  _ <
 * |_|  |_|___/_/\_\_____|_| \_\
 * 
 */
static void append_sourcetime_to_buffer(gpointer el, gpointer buf)
{
	GdamSoundSpecMixElement* mix_element = (GdamSoundSpecMixElement*)el;
	GdamBuffer* buffer = (GdamBuffer*)buf;

	gdam_buffer_append_time(buffer, mix_element->start_time);
	mix_element->spec->any_spec.to_buffer(mix_element->spec, buffer);
}
static void gdam_sound_spec_mix_to_buffer(GdamSoundSpec* spec,
                                          GdamBuffer*    into)
{
	GdamBuffer rv;
	GdamSoundSpecMix* mix = &spec->mix_spec;
	g_assert(spec->type == GDAM_SOURCE_MIX);
	gdam_buffer_construct(&rv);
	gdam_buffer_append_int(&rv, GDAM_SOURCE_MIX);
	gdam_buffer_append_int(&rv, mix->terminates);
	gdam_buffer_append_int(&rv, g_list_length(mix->first));
	g_list_foreach(mix->first, append_sourcetime_to_buffer, &rv);
	//gdam_buffer_prepend_length(&rv);
	gdam_buffer_append_buffer(into, &rv, TRUE);
}

static void gdam_sound_spec_mix_pretty_print(GdamSoundSpec*  spec,
	                                     GdamPrinter*    printer)
{
	GdamSoundSpecMix* mix = &spec->mix_spec;
	int index = 0;
	GList* list;
	gdam_printer_write_line(printer, "MIX: %s, %d subsources.",
			(mix->terminates ? "temporary" : "eternal"),
			g_list_length(mix->first));
	gdam_printer_indent(printer);
	for (list = mix->first; list; list = list->next) {
		GdamSoundSpecMixElement* mix_element;
		GdamSoundSpecAny* any;
		mix_element = (GdamSoundSpecMixElement*)(list->data);
		any = &mix_element->spec->any_spec;
		gdam_printer_write_line(printer, "Source %d, time %lld",
				index, mix_element->start_time);
		gdam_printer_indent(printer);
		any->pretty_print(mix_element->spec, printer);
		gdam_printer_unindent(printer);
		index++;
	}
	gdam_printer_unindent(printer);
}

static void destroy_element(gpointer el, gpointer unused)
{
	GdamSoundSpecMixElement* mix_element = (GdamSoundSpecMixElement*)el;
	mix_element->spec->any_spec.destroy(mix_element->spec);
	g_free(mix_element);
}

static void gdam_sound_spec_mix_destroy(GdamSoundSpec* spec)
{
	GdamSoundSpecMix* mix = &spec->mix_spec;
	g_assert(spec->type == GDAM_SOURCE_MIX);
	g_list_foreach(mix->first, destroy_element, NULL);
	g_list_free(mix->first);
	g_free(spec);
}

GdamSoundSpec* gdam_sound_spec_mix_list(int           terminates,
				        GList*	      elements)
{
	GdamSoundSpecMix* rv = g_new(GdamSoundSpecMix, 1);
	rv->base.type = GDAM_SOURCE_MIX;
	rv->base.destroy = gdam_sound_spec_mix_destroy;
	rv->base.to_buffer = gdam_sound_spec_mix_to_buffer;
	rv->base.pretty_print = gdam_sound_spec_mix_pretty_print;
	rv->first = elements;
	rv->terminates = terminates;
	rv->last = g_list_last(elements);
	return (GdamSoundSpec*) rv;
}

GdamSoundSpec* gdam_sound_spec_mix  (int            terminates,
                                     int            num_subs,
				     ...)
{
	GList* list = NULL;
	va_list args;
	va_start(args, num_subs);
	while (num_subs--) {
		GdamSoundSpecMixElement* element;
		element = g_new(GdamSoundSpecMixElement, 1);
		element->start_time = va_arg(args, GdamTime);
		element->spec = va_arg(args, GdamSoundSpec*);
		list = g_list_prepend(list, element);
	}
	list = g_list_reverse(list);

	return gdam_sound_spec_mix_list(terminates, list);
}

/*
 *  __  __ ____ _____
 * |  \/  |  _ \___ /
 * | |\/| | |_) ||_ \
 * | |  | |  __/___) |
 * |_|  |_|_|  |____/
 * 
 */
static void gdam_sound_spec_mp3_destroy(GdamSoundSpec* spec)
{
	GdamSoundSpecMp3* mp3 = &spec->mp3_spec;
	g_assert(spec->type == GDAM_SOURCE_MP3);
	g_free(mp3->mp3_filename);
	g_free(spec);
}

static void gdam_sound_spec_mp3_to_buffer(GdamSoundSpec* spec,
                                          GdamBuffer* buffer)
{
	GdamSoundSpecMp3* mp3 = &spec->mp3_spec;
	GdamBuffer tmp;
	g_assert(spec->type == GDAM_SOURCE_MP3);
	gdam_buffer_construct(&tmp);
	gdam_buffer_append_int(&tmp, GDAM_SOURCE_MP3);
	gdam_buffer_append_string0(&tmp, mp3->mp3_filename);
	gdam_buffer_append_time(&tmp, mp3->start);
	gdam_buffer_append_time(&tmp, mp3->length);
	//gdam_buffer_prepend_length(&tmp);
	gdam_buffer_append_buffer(buffer, &tmp, TRUE);
}
static void gdam_sound_spec_mp3_pretty_print(GdamSoundSpec*  spec,
	                                     GdamPrinter*    printer)
{
	GdamSoundSpecMp3* mp3 = &spec->mp3_spec;
	gdam_printer_write_line(printer,
	                        "MP3: %s. position %lld, length %lld",
	                        mp3->mp3_filename,
	                        mp3->start,
	                        mp3->length);
}

GdamSoundSpec* gdam_sound_spec_mp3    (const char*    filename,
                                       GdamTime       start_time,
				       GdamTime       length)
{
	GdamSoundSpecMp3* rv = g_new(GdamSoundSpecMp3, 1);
	rv->base.type = GDAM_SOURCE_MP3;
	rv->base.destroy = gdam_sound_spec_mp3_destroy;
	rv->base.to_buffer = gdam_sound_spec_mp3_to_buffer;
	rv->base.pretty_print = gdam_sound_spec_mp3_pretty_print;
	rv->mp3_filename = g_strdup(filename);
	rv->start = start_time;
	rv->length = length;
	return (GdamSoundSpec*) rv;
}

/*
 *              _
 *   __ _  __ _(_)_ __
 *  / _` |/ _` | | '_ \
 * | (_| | (_| | | | | |
 *  \__, |\__,_|_|_| |_|
 *  |___/
 */
static void gdam_sound_spec_gain_destroy(GdamSoundSpec* spec)
{
	GdamSoundSpecGain* gain = &spec->gain_spec;
	g_assert(spec->type == GDAM_SOURCE_GAIN);
	if (gain->subsource)
		gdam_sound_spec_destroy(gain->subsource);
	g_free(spec);
}

static void gdam_sound_spec_gain_to_buffer(GdamSoundSpec* spec,
                                           GdamBuffer* buffer)
{
	GdamSoundSpecGain* gain = &spec->gain_spec;
	GdamBuffer tmp;
	g_assert(spec->type == GDAM_SOURCE_GAIN);
	gdam_buffer_construct(&tmp);
	gdam_buffer_append_int(&tmp, GDAM_SOURCE_GAIN);
	gdam_buffer_append_int(&tmp, gain->level);
	if (gain->subsource) {
		gdam_buffer_append_int(&tmp, 1);
		gdam_sound_spec_write(gain->subsource, &tmp);
	} else
		gdam_buffer_append_int(&tmp, 0);
	gdam_buffer_append_buffer(buffer, &tmp, TRUE);
}
static void gdam_sound_spec_gain_pretty_print(GdamSoundSpec*  spec,
	                                      GdamPrinter*    printer)
{
	GdamSoundSpecGain* gain = &spec->gain_spec;
	gdam_printer_write_line(printer, "GAIN: volume %f.",
	                       (gdouble)(gain->level) / GdamFixed_ONE);
	gdam_printer_indent(printer);
	gdam_sound_spec_pretty_print(gain->subsource, printer);
	gdam_printer_unindent(printer);
}

GdamSoundSpec* gdam_sound_spec_gain   (int            level,
                                       GdamSoundSpec* spec)
{
	GdamSoundSpecGain* rv = g_new(GdamSoundSpecGain, 1);
	rv->base.type = GDAM_SOURCE_GAIN;
	rv->base.destroy = gdam_sound_spec_gain_destroy;
	rv->base.to_buffer = gdam_sound_spec_gain_to_buffer;
	rv->base.pretty_print = gdam_sound_spec_gain_pretty_print;
	rv->level = level;
	rv->subsource = spec;
	return (GdamSoundSpec*) rv;
}

/*
 *                                      _
 *  _ __ ___  ___  __ _ _ __ ___  _ __ | | ___
 * | '__/ _ \/ __|/ _` | '_ ` _ \| '_ \| |/ _ \
 * | | |  __/\__ \ (_| | | | | | | |_) | |  __/
 * |_|  \___||___/\__,_|_| |_| |_| .__/|_|\___|
 *                               |_|
 */
static void gdam_sound_spec_resample_destroy(GdamSoundSpec* spec)
{
	GdamSoundSpecResample* resample = &spec->resample_spec;
	g_assert(spec->type == GDAM_SOURCE_RESAMPLE);
	if (resample->subsource)
		gdam_sound_spec_destroy(resample->subsource);
	g_free(spec);
}

static void gdam_sound_spec_resample_to_buffer(GdamSoundSpec* spec,
                                           GdamBuffer* buffer)
{
	GdamSoundSpecResample* resample = &spec->resample_spec;
	GdamBuffer tmp;
	g_assert(spec->type == GDAM_SOURCE_RESAMPLE);
	gdam_buffer_construct(&tmp);
	gdam_buffer_append_int(&tmp, GDAM_SOURCE_RESAMPLE);
	gdam_buffer_append_int(&tmp, resample->rate);
	if (resample->subsource) {
		gdam_buffer_append_int(&tmp, 1);
		gdam_sound_spec_write(resample->subsource, &tmp);
	} else
		gdam_buffer_append_int(&tmp, 0);
	gdam_buffer_append_buffer(buffer, &tmp, TRUE);
}

static void gdam_sound_spec_resample_pretty_print(GdamSoundSpec*  spec,
	                                      GdamPrinter*    printer)
{
	GdamSoundSpecResample* resample = &spec->resample_spec;
	gdam_printer_write_line(printer, "RESAMPLE: rate %f.",
	                       (gdouble)(resample->rate) / GdamFixed_ONE);
	gdam_printer_indent(printer);
	gdam_sound_spec_pretty_print(resample->subsource, printer);
	gdam_printer_unindent(printer);
}


GdamSoundSpec* gdam_sound_spec_resample(int            rate,
                                        GdamSoundSpec* spec)
{
	GdamSoundSpecResample* rv = g_new(GdamSoundSpecResample, 1);
	rv->base.type = GDAM_SOURCE_RESAMPLE;
	rv->base.destroy = gdam_sound_spec_resample_destroy;
	rv->base.to_buffer = gdam_sound_spec_resample_to_buffer;
	rv->base.pretty_print = gdam_sound_spec_resample_pretty_print;
	rv->rate = rate;
	rv->subsource = spec;
	return (GdamSoundSpec*) rv;
}


/*
 *            _
 *   ___  ___| |__   ___
 *  / _ \/ __| '_ \ / _ \
 * |  __/ (__| | | | (_) |
 *  \___|\___|_| |_|\___/
 * 
 */
static void gdam_sound_spec_echo_destroy(GdamSoundSpec* spec)
{
	GdamSoundSpecEcho* echo = &spec->echo_spec;
	g_assert(spec->type == GDAM_SOURCE_ECHO);
	if (echo->subsource)
		gdam_sound_spec_destroy(echo->subsource);
	if (echo->echoes)
		g_free(echo->echoes);
	g_free(spec);
}

static void gdam_sound_spec_echo_to_buffer(GdamSoundSpec* spec,
                                           GdamBuffer* buffer)
{
	GdamSoundSpecEcho* echo = &spec->echo_spec;
	GdamBuffer tmp;
	int which_echo;
	g_assert(spec->type == GDAM_SOURCE_ECHO);
	gdam_buffer_construct(&tmp);
	gdam_buffer_append_int(&tmp, GDAM_SOURCE_ECHO);
	gdam_buffer_append_int(&tmp, echo->reverb ? 1 : 0);
	gdam_buffer_append_int(&tmp, echo->num_echoes);
	for (which_echo = 0; which_echo < echo->num_echoes; which_echo++) {
		gdam_buffer_append_int(&tmp, echo->echoes[which_echo].volume);
		gdam_buffer_append_time(&tmp, echo->echoes[which_echo].delay);
	}
	if (echo->subsource) {
		gdam_buffer_append_int(&tmp, 1);
		gdam_sound_spec_write(echo->subsource, &tmp);
	} else
		gdam_buffer_append_int(&tmp, 0);
	gdam_buffer_append_buffer(buffer, &tmp, TRUE);
}

static void gdam_sound_spec_echo_pretty_print(GdamSoundSpec*  spec,
	                                      GdamPrinter*    printer)
{
	GdamSoundSpecEcho* echo = &spec->echo_spec;
	int i;
	gdam_printer_write_line(printer, "ECHO: %d echoes.",
				echo->num_echoes);
	for (i = 0; i < echo->num_echoes; i++) {
		gdam_printer_write_line(
			printer,
			"- echo %d, volume %2.4f, delay %2.4fs.",
			i,
			(float)echo->echoes[i].volume / (float)GdamFixed_ONE,
			(float)echo->echoes[i].delay / (float)44100.0
		);
	}
	gdam_printer_indent(printer);
	gdam_sound_spec_pretty_print(echo->subsource, printer);
	gdam_printer_unindent(printer);
}

GdamSoundSpec* gdam_sound_spec_echo_v (GdamSoundSpec* spec,
                                       gboolean       reverb,
                                       int            num_echoes,
				       const GdamSoundSpecEchoElement* echoes)
{
	GdamSoundSpecEcho* rv = g_new(GdamSoundSpecEcho, 1);
	rv->base.type = GDAM_SOURCE_ECHO;
	rv->base.destroy = gdam_sound_spec_echo_destroy;
	rv->base.to_buffer = gdam_sound_spec_echo_to_buffer;
	rv->base.pretty_print = gdam_sound_spec_echo_pretty_print;
	rv->subsource = spec;
	rv->num_echoes = num_echoes;
	rv->echoes = g_new(GdamSoundSpecEchoElement, num_echoes);
	rv->reverb = reverb ? 1 : 0;
	memcpy(rv->echoes, echoes,
	       sizeof(GdamSoundSpecEchoElement) * num_echoes);
	return (GdamSoundSpec*) rv;
}

GdamSoundSpec* gdam_sound_spec_echo   (GdamSoundSpec* sub,
                                       gboolean       reverb,
                                       int            num_echoes,
				       ...);	/* volume/delay pairs... */

/*
 *                 _
 *   ___ __ _  ___| |__   ___
 *  / __/ _` |/ __| '_ \ / _ \
 * | (_| (_| | (__| | | |  __/
 *  \___\__,_|\___|_| |_|\___|
 * 
 */
static void gdam_sound_spec_cache_destroy(GdamSoundSpec* spec)
{
	//GdamSoundSpecById* cache = &spec->cache_spec;
	g_assert(spec->type == GDAM_SOURCE_CACHE);
	g_free(spec);
}

static void gdam_sound_spec_cache_to_buffer(GdamSoundSpec* spec,
                                           GdamBuffer* buffer)
{
	GdamSoundSpecCache* cache = &spec->cache_spec;
	g_assert(spec->type == GDAM_SOURCE_CACHE);
	gdam_buffer_append_int(buffer, GDAM_SOURCE_CACHE);
	gdam_buffer_append_int(buffer, cache->buffer_id);
	gdam_buffer_append_int(buffer, cache->was_loop ? 1 : 0);
	gdam_buffer_append_int(buffer, cache->start_sample);
}

static void gdam_sound_spec_cache_pretty_print(GdamSoundSpec*  spec,
	                                      GdamPrinter*    printer)
{
	GdamSoundSpecCache* cache = &spec->cache_spec;
	gdam_printer_write_line(printer,
			"CACHE: buffer_id=%d, position=%d, looping=%s",
					cache->buffer_id,
					cache->start_sample,
					(cache->was_loop ? "yes" : "no"));
}

GdamSoundSpec* gdam_sound_spec_cache  (int            buffer_id,
                                       gboolean       was_loop,
				       int            start_sample)
{
	GdamSoundSpecCache* rv = g_new(GdamSoundSpecCache, 1);
	rv->base.type = GDAM_SOURCE_CACHE;
	rv->base.destroy = gdam_sound_spec_cache_destroy;
	rv->base.to_buffer = gdam_sound_spec_cache_to_buffer;
	rv->base.pretty_print = gdam_sound_spec_cache_pretty_print;
	rv->buffer_id = buffer_id;
	rv->start_sample = start_sample;
	rv->was_loop = was_loop;
	return (GdamSoundSpec*) rv;
}

/*
 *                          _        _     _
 * __      ____ ___   _____| |_ __ _| |__ | | ___
 * \ \ /\ / / _` \ \ / / _ \ __/ _` | '_ \| |/ _ \
 *  \ V  V / (_| |\ V /  __/ || (_| | |_) | |  __/
 *   \_/\_/ \__,_| \_/ \___|\__\__,_|_.__/|_|\___|
 * 
 */
static void gdam_sound_spec_wavetable_destroy(GdamSoundSpec* spec)
{
	//GdamSoundSpecById* wavetable = &spec->wavetable_spec;
	g_assert(spec->type == GDAM_SOURCE_WAVETABLE);
	g_free(spec);
}

static void gdam_sound_spec_wavetable_to_buffer(GdamSoundSpec* spec,
                                           GdamBuffer* buffer)
{
	GdamSoundSpecWavetable* wavetable = &spec->wavetable_spec;
	gint hi, lo;
	g_assert(spec->type == GDAM_SOURCE_WAVETABLE);
	gdam_buffer_append_int(buffer, GDAM_SOURCE_WAVETABLE);
	gdam_buffer_append_int(buffer, wavetable->buffer_id);
	hi = (gint32)(wavetable->hertz_32>>32);
	lo = (gint32)(wavetable->hertz_32);
	gdam_buffer_append_int(buffer, hi);
	gdam_buffer_append_int(buffer, lo);
}

static void gdam_sound_spec_wavetable_pretty_print(GdamSoundSpec*  spec,
	                                      GdamPrinter*    printer)
{
	GdamSoundSpecWavetable* wavetable = &spec->wavetable_spec;
	gdam_printer_write_line(printer,
			"WAVETABLE: buffer_id=%d, hertz=%3.3g",
					wavetable->buffer_id,
					(gdouble)wavetable->hertz_32 /
					         (double)(1LL<<32));
}

GdamSoundSpec* gdam_sound_spec_wavetable(int            buffer_id,
                                         gint64         hertz_32)
{
	GdamSoundSpecWavetable* rv = g_new(GdamSoundSpecWavetable, 1);
	rv->base.type = GDAM_SOURCE_WAVETABLE;
	rv->base.destroy = gdam_sound_spec_wavetable_destroy;
	rv->base.to_buffer = gdam_sound_spec_wavetable_to_buffer;
	rv->base.pretty_print = gdam_sound_spec_wavetable_pretty_print;
	rv->buffer_id = buffer_id;
	rv->hertz_32 = hertz_32;
	return (GdamSoundSpec*) rv;
}

/*
 * __        _____     __
 * \ \      / / \ \   / /
 *  \ \ /\ / / _ \ \ / /
 *   \ V  V / ___ \ V /
 *    \_/\_/_/   \_\_/
 * 
 */
static void gdam_sound_spec_wav_destroy(GdamSoundSpec* spec)
{
	GdamSoundSpecWav* wav = &spec->wav_spec;
	g_assert(spec->type == GDAM_SOURCE_WAV);
	g_free(wav->wav_filename);
	g_free(spec);
}

static void gdam_sound_spec_wav_to_buffer(GdamSoundSpec* spec,
                                          GdamBuffer* buffer)
{
	GdamSoundSpecWav* wav = &spec->wav_spec;
	GdamBuffer tmp;
	g_assert(spec->type == GDAM_SOURCE_WAV);
	gdam_buffer_construct(&tmp);
	gdam_buffer_append_int(&tmp, GDAM_SOURCE_WAV);
	gdam_buffer_append_string0(&tmp, wav->wav_filename);
	//gdam_buffer_prepend_length(&tmp);
	gdam_buffer_append_buffer(buffer, &tmp, TRUE);
}
static void gdam_sound_spec_wav_pretty_print(GdamSoundSpec*  spec,
	                                     GdamPrinter*    printer)
{
	GdamSoundSpecWav* wav = &spec->wav_spec;
	gdam_printer_write_line(printer, "WAV: %s.", wav->wav_filename);
}

GdamSoundSpec* gdam_sound_spec_wav    (const char*    filename)
{
	GdamSoundSpecWav* rv = g_new(GdamSoundSpecWav, 1);
	rv->base.type = GDAM_SOURCE_WAV;
	rv->base.destroy = gdam_sound_spec_wav_destroy;
	rv->base.to_buffer = gdam_sound_spec_wav_to_buffer;
	rv->base.pretty_print = gdam_sound_spec_wav_pretty_print;
	rv->wav_filename = g_strdup(filename);
	return (GdamSoundSpec*) rv;
}


/*
 *  _             _     _
 * | |__  _   _  (_) __| |
 * | '_ \| | | | | |/ _` |
 * | |_) | |_| | | | (_| |
 * |_.__/ \__, | |_|\__,_|
 *        |___/
 */
static void gdam_sound_spec_by_id_destroy(GdamSoundSpec* spec)
{
	//GdamSoundSpecById* by_id = &spec->by_id_spec;
	g_assert(spec->type == GDAM_SOURCE_BY_ID);
	g_free(spec);
}

static void gdam_sound_spec_by_id_to_buffer(GdamSoundSpec* spec,
                                           GdamBuffer* buffer)
{
	GdamSoundSpecById* by_id = &spec->by_id_spec;
	g_assert(spec->type == GDAM_SOURCE_BY_ID);
	gdam_buffer_append_int(buffer, GDAM_SOURCE_BY_ID);
	gdam_buffer_append_int(buffer, by_id->source_id);
}
static void gdam_sound_spec_by_id_pretty_print(GdamSoundSpec*  spec,
	                                      GdamPrinter*    printer)
{
	GdamSoundSpecById* by_id = &spec->by_id_spec;
	gdam_printer_write_line(printer, "BY_ID: %d", by_id->source_id);
}

GdamSoundSpec* gdam_sound_spec_by_id (int source_id)
{
	GdamSoundSpecById* rv = g_new(GdamSoundSpecById, 1);
	rv->base.type = GDAM_SOURCE_BY_ID;
	rv->base.destroy = gdam_sound_spec_by_id_destroy;
	rv->base.to_buffer = gdam_sound_spec_by_id_to_buffer;
	rv->base.pretty_print = gdam_sound_spec_by_id_pretty_print;
	rv->source_id = source_id;
	return (GdamSoundSpec*) rv;
}

/*
 * 
 *  ___  ___  __ _ _   _  ___ _ __   ___ ___
 * / __|/ _ \/ _` | | | |/ _ \ '_ \ / __/ _ \
 * \__ \  __/ (_| | |_| |  __/ | | | (_|  __/
 * |___/\___|\__, |\__,_|\___|_| |_|\___\___|
 *              |_|
 */


static void gdam_sound_spec_sequencer_destroy(GdamSoundSpec* spec)
{
	//GdamSoundSpecSequence* sequencer = &spec->sequencer_spec;
	g_assert(spec->type == GDAM_SOURCE_SEQUENCER);
	g_free(spec);
}

static void gdam_sound_spec_sequencer_to_buffer(GdamSoundSpec* spec,
                                           GdamBuffer* buffer)
{
	int i;
	GdamSoundSpecSequencer* sequencer = &spec->sequencer_spec;
	g_assert(spec->type == GDAM_SOURCE_SEQUENCER);
	gdam_buffer_append_int(buffer, GDAM_SOURCE_SEQUENCER);
	gdam_buffer_append_int(buffer, sequencer->total_time);
	gdam_buffer_append_int(buffer, sequencer->num_elements);
	for (i = 0; i < sequencer->num_elements; i++) {
		GdamSoundSpecSequencerElement* element;
		element = sequencer->elements + i;
		gdam_buffer_append_int(buffer, element->buffer_id);
		gdam_buffer_append_int(buffer, element->start_offset);
		gdam_buffer_append_int(buffer, element->resample_b16);
		gdam_buffer_append_int(buffer, element->gain_b16);
		gdam_buffer_append_int(buffer, element->sample_offset);
		gdam_buffer_append_int(buffer, element->output_length);
	}
}
static void gdam_sound_spec_sequencer_pretty_print(GdamSoundSpec*  spec,
	                                      GdamPrinter*    printer)
{
	GdamSoundSpecSequencer* sequencer = &spec->sequencer_spec;
	int i;
	gdam_printer_write_line(printer, "SEQUENCE: length %d, num elements %d",
			sequencer->total_time, sequencer->num_elements);
	gdam_printer_indent(printer);
	for (i = 0; i < sequencer->num_elements; i++) {
		GdamSoundSpecSequencerElement* el;
		el = sequencer->elements + i;
		gdam_printer_write_line(printer, "Note %d: buffer %d, "
					"speed %.3f, volume %.3f, startpos %d,"
					" sampleoffset %d, length %d",
					i, el->buffer_id,
				(float)(el->resample_b16) / GdamFixed_ONE,
				(float)(el->gain_b16) / GdamFixed_ONE,
				(int)el->start_offset, (int)el->sample_offset,
				(int)el->output_length
				);
	}
	gdam_printer_unindent(printer);
}

GdamSoundSpec* gdam_sound_spec_sequencer (int sequencer_length,
                                 int num_sub,
				 const GdamSoundSpecSequencerElement* els)
{
	GdamSoundSpecSequencer* rv = g_new(GdamSoundSpecSequencer, 1);
	rv->base.type = GDAM_SOURCE_SEQUENCER;
	rv->base.destroy = gdam_sound_spec_sequencer_destroy;
	rv->base.to_buffer = gdam_sound_spec_sequencer_to_buffer;
	rv->base.pretty_print = gdam_sound_spec_sequencer_pretty_print;
	rv->total_time = sequencer_length;
	rv->num_elements = num_sub;
	rv->elements = g_new(GdamSoundSpecSequencerElement, num_sub);
	memcpy(rv->elements,
	       els,
	       num_sub * sizeof(GdamSoundSpecSequencerElement));
	return (GdamSoundSpec*) rv;
}




/*
 *                      _       _       _
 *  _ __ ___   ___   __| |_   _| | __ _| |_ ___  _ __
 * | '_ ` _ \ / _ \ / _` | | | | |/ _` | __/ _ \| '__|
 * | | | | | | (_) | (_| | |_| | | (_| | || (_) | |
 * |_| |_| |_|\___/ \__,_|\__,_|_|\__,_|\__\___/|_|
 * 
 */
static void gdam_sound_spec_modulator_to_buffer(GdamSoundSpec* spec,
                                          GdamBuffer*    into)
{
	GdamBuffer rv;
	GdamSoundSpecModulator* modulator = &spec->modulator_spec;
	g_assert(spec->type == GDAM_SOURCE_MODULATOR);
	gdam_buffer_construct(&rv);
	gdam_buffer_append_int(&rv, GDAM_SOURCE_MODULATOR);
	gdam_buffer_append_int(&rv, modulator->terminates);
	gdam_buffer_append_int(&rv, g_list_length(modulator->first));
	g_list_foreach(modulator->first, append_sourcetime_to_buffer, &rv);
	gdam_buffer_append_buffer(into, &rv, TRUE);
}

static void gdam_sound_spec_modulator_pretty_print(GdamSoundSpec*  spec,
	                                     GdamPrinter*    printer)
{
	GdamSoundSpecModulator* modulator = &spec->modulator_spec;
	int index = 0;
	GList* list;
	gdam_printer_write_line(printer, "MODULATOR: %s, %d subsources.",
			(modulator->terminates ? "temporary" : "eternal"),
			g_list_length(modulator->first));
	gdam_printer_indent(printer);
	for (list = modulator->first; list; list = list->next) {
		GdamSoundSpecMixElement* modulator_element;
		GdamSoundSpecAny* any;
		modulator_element = (GdamSoundSpecMixElement*)(list->data);
		any = &modulator_element->spec->any_spec;
		gdam_printer_write_line(printer, "Source %d, time %lld",
				index, modulator_element->start_time);
		gdam_printer_indent(printer);
		any->pretty_print(modulator_element->spec, printer);
		gdam_printer_unindent(printer);
		index++;
	}
	gdam_printer_unindent(printer);
}

static void gdam_sound_spec_modulator_destroy(GdamSoundSpec* spec)
{
	GdamSoundSpecModulator* modulator = &spec->modulator_spec;
	g_assert(spec->type == GDAM_SOURCE_MODULATOR);
	g_list_foreach(modulator->first, destroy_element, NULL);
	g_list_free(modulator->first);
	g_free(spec);
}

GdamSoundSpec* gdam_sound_spec_modulator_list(int           terminates,
				        GList*	      elements)
{
	GdamSoundSpecModulator* rv = g_new(GdamSoundSpecModulator, 1);
	rv->base.type = GDAM_SOURCE_MODULATOR;
	rv->base.destroy = gdam_sound_spec_modulator_destroy;
	rv->base.to_buffer = gdam_sound_spec_modulator_to_buffer;
	rv->base.pretty_print = gdam_sound_spec_modulator_pretty_print;
	rv->first = elements;
	rv->terminates = terminates;
	rv->last = g_list_last(elements);
	return (GdamSoundSpec*) rv;
}

GdamSoundSpec* gdam_sound_spec_modulator  (int            terminates,
                                     int            num_subs,
				     ...)
{
	GList* list = NULL;
	va_list args;
	va_start(args, num_subs);
	while (num_subs--) {
		GdamSoundSpecMixElement* element;
		element = g_new(GdamSoundSpecMixElement, 1);
		element->start_time = va_arg(args, GdamTime);
		element->spec = va_arg(args, GdamSoundSpec*);
		list = g_list_prepend(list, element);
	}
	list = g_list_reverse(list);

	return gdam_sound_spec_modulator_list(terminates, list);
}

/*
 *                _ _
 * __      ____ _| | | _____ _ __
 * \ \ /\ / / _` | | |/ / _ \ '__|
 *  \ V  V / (_| | |   <  __/ |
 *   \_/\_/ \__,_|_|_|\_\___|_|
 * 
 */
static void gdam_sound_spec_walker_destroy(GdamSoundSpec* spec)
{
	//GdamSoundSpecById* walker = &spec->walker_spec;
	g_assert(spec->type == GDAM_SOURCE_WALKER);
	g_free(spec);
}

static void gdam_sound_spec_walker_to_buffer(GdamSoundSpec* spec,
                                           GdamBuffer* buffer)
{
	GdamSoundSpecWalker* walker = &spec->walker_spec;
	g_assert(spec->type == GDAM_SOURCE_WALKER);
	gdam_buffer_append_int(buffer, GDAM_SOURCE_WALKER);
	gdam_buffer_append_int(buffer, walker->buffer_id);
}

static void gdam_sound_spec_walker_pretty_print(GdamSoundSpec*  spec,
	                                      GdamPrinter*    printer)
{
	GdamSoundSpecWalker* walker = &spec->walker_spec;
	gdam_printer_write_line(printer,
			"WALKER: buffer_id=%d", walker->buffer_id);
}

GdamSoundSpec* gdam_sound_spec_walker(int            buffer_id)
{
	GdamSoundSpecWalker* rv = g_new(GdamSoundSpecWalker, 1);
	rv->base.type = GDAM_SOURCE_WALKER;
	rv->base.destroy = gdam_sound_spec_walker_destroy;
	rv->base.to_buffer = gdam_sound_spec_walker_to_buffer;
	rv->base.pretty_print = gdam_sound_spec_walker_pretty_print;
	rv->buffer_id = buffer_id;
	return (GdamSoundSpec*) rv;
}


/*
 *  _   _  ___  _   _            _      _               _
 * | \ | |/ _ \| \ | |    __   _(_)_ __| |_ _   _  __ _| |___
 * |  \| | | | |  \| |____\ \ / / | '__| __| | | |/ _` | / __|
 * | |\  | |_| | |\  |_____\ V /| | |  | |_| |_| | (_| | \__ \
 * |_| \_|\___/|_| \_|      \_/ |_|_|   \__|\__,_|\__,_|_|___/
 * 
 */

void gdam_sound_spec_pretty_print(GdamSoundSpec* spec, GdamPrinter* printer)
{
	if (spec == NULL) {
		gdam_printer_write_line(printer, "(null)");
		return;
	}
	g_assert(spec->any_spec.pretty_print != NULL);
	spec->any_spec.pretty_print(spec, printer);
}

void gdam_sound_spec_write(GdamSoundSpec* spec,
                           GdamBuffer*    output_buffer)
{
	spec->any_spec.to_buffer(spec, output_buffer);
}

void gdam_sound_spec_destroy(GdamSoundSpec* spec)
{
	spec->any_spec.destroy(spec);
}

