/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamsequencer.h"
#include <gdam/gdamsample.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkmarshal.h>
#include "gdambuffermanager.h"
#include <gdam/component/gdamutil.h>
#include <math.h>

enum {
	SET_NUM_BEATS,
	SEQUENCE_CHANGED,
	LAST_SIGNAL
};
static guint sequencer_signals[LAST_SIGNAL] = {0};

typedef struct _GdamSequencerRescaleInfo GdamSequencerRescaleInfo;

struct _GdamSequencerRescaleInfo {
	GdamSequencer* 	sequencer;
	double		rescale_factor;
};

static gboolean gdam_sequencer_match_beat_info(GdamModel*	    model,
				  	       const GdamBeatInfo*  new_info);
static gboolean gdam_sequencer_set_beat_info  (GdamModel*           model,
	                                       const GdamBeatInfo*  new_info);
static void gdam_sequencer_detach_from_buffer(GdamModel*     model,
				              guint          buffer_id);

static inline gdouble beat_to_samples(GdamSequencer* sequencer, gdouble beat)
{
	gdouble bpm = 
		gdam_beat_info_get_bpm(&GDAM_MODEL(sequencer)->beat_info,0);
	return (60.0 / bpm * beat * 44100.0);
}

static inline gdouble samples_to_beats(GdamSequencer* sequencer, gdouble pos)
{
	gdouble bpm = 
		gdam_beat_info_get_bpm(&GDAM_MODEL(sequencer)->beat_info,0);
	return (bpm / 60.0 / 44100.0) * pos;
}

static inline gdouble seconds_to_beats(GdamSequencer* sequencer, gdouble pos)
{
	gdouble bpm = 
		gdam_beat_info_get_bpm(&GDAM_MODEL(sequencer)->beat_info,0);
	return (bpm / 60.0) * pos;
}

static gint* gdam_sequencer_create_array(GdamSequencer* sequencer,
                                         int*           array_length_out)
{
	int num_elements = g_list_length(sequencer->elements);
	gint* rv = g_new(gint, num_elements * 6);
	int i;
	GList* at = sequencer->elements;
	gdouble beat_length = beat_to_samples(sequencer, 1.0);
	for (i = 0; at; i++, at = at->next) {
		GdamSequencerElement* element = at->data;
		rv[i*6+0] = element->buffer_id;
		rv[i*6+1] = (gint)(element->beat_offset * beat_length);
		rv[i*6+2] = (gint)(element->rate * GdamFixed_ONE);
		rv[i*6+3] = (gint)(element->gain * GdamFixed_ONE);
		rv[i*6+4] = element->sample_offset;
		rv[i*6+5] = element->output_length;
	}
	*array_length_out = num_elements * 6;
	return rv;
}


void gdam_sequencer_update_sequence(GdamSequencer* sequencer)
{
	gint* array;
	int array_length;
	GdamRequest* request;
	GdamModel* this_model = GDAM_MODEL(sequencer);
	GdamArg* arg;

	if (sequencer->frozen)
		return;
	if (this_model->source_id == 0)
		return;

	array = gdam_sequencer_create_array(sequencer, &array_length);
	arg = gdam_arg_new_from_valist(1, 
	                               "elements",
				       GDAM_ARG_TYPE_INT_ARRAY,
				       array_length,
				       array);
	request = gdam_request_set_args(this_model->source_id,
	                                1,
					arg,
					NULL,
					NULL,
					NULL);
	gdam_channel_append_request(this_model->channel, request);
}

/* TODO: start within the sequence */
static void gdam_sequencer_play(GdamProducer* producer, 
				GdamPlayInfo* play_info)
{
	GdamSequencer* sequencer = GDAM_SEQUENCER(producer);
	GdamRequest* request;
	GdamRequestBuilder* request_builder;
	gint* array;
	int array_length;
	int new_id;
	int sequence_length;
	g_return_if_fail(sequencer != NULL);
	if (GDAM_MODEL(producer)->source_id != 0) 
		return;
	sequence_length = (int) beat_to_samples(sequencer,
	                                        sequencer->num_beats);
	array = gdam_sequencer_create_array(sequencer, &array_length);
	gtk_object_ref(GTK_OBJECT(sequencer));
	GDAM_MODEL(producer)->create_source_pending = 1;
	request = gdam_request_create_source("GdamSequencer",
					     gdam_model_got_subsource,
					     gdam_model_didnt_get_subsource,
					     sequencer,
	                                     2,
					     "sequence_length",
					     GDAM_ARG_TYPE_INT,
					     sequence_length,
					     "elements",
					     GDAM_ARG_TYPE_INT_ARRAY,
					     array_length,
					     array);
	g_free(array);
	sequencer->making_request = 1;
	request_builder = gdam_request_builder_new();
	new_id = gdam_request_builder_append_create(request_builder,
	                                            request);
	gdam_model_new_child_request(GDAM_MODEL(sequencer)->parent,
	                                 GDAM_MODEL(sequencer),
					 play_info,
	                                 new_id,
					 request_builder);
}

static void gdam_sequencer_init(GdamSequencer* sequencer)
{
	sequencer->num_beats 	   = 4;
	sequencer->num_sub_beats   = 4;
	gdam_beat_info_set_bpm(&GDAM_MODEL(sequencer)->beat_info, 60.0);
}
static void gdam_sequencer_class_init(GdamSequencerClass* sequencer_class)
{
	GdamModelClass* model_class;
	GtkObjectClass* object_class = GTK_OBJECT_CLASS(sequencer_class);
	model_class = GDAM_MODEL_CLASS(sequencer_class);
	sequencer_class->producer_class.play = gdam_sequencer_play;
	model_class->match_beat_info = gdam_sequencer_match_beat_info;
	model_class->set_beat_info = gdam_sequencer_set_beat_info;
	model_class->detach_from_buffer = gdam_sequencer_detach_from_buffer;
	sequencer_signals[SET_NUM_BEATS] =
		gtk_signal_new(
			"set_num_beats",
			GTK_RUN_LAST,
			object_class->type,
			GTK_SIGNAL_OFFSET(GdamSequencerClass, set_num_beats),
			gtk_marshal_NONE__DOUBLE,
			GTK_TYPE_NONE,
			1,
			GTK_TYPE_DOUBLE
		); 
	sequencer_signals[SEQUENCE_CHANGED] =
		gtk_signal_new(
			"sequence_changed",
			GTK_RUN_LAST,
			object_class->type,
			GTK_SIGNAL_OFFSET(GdamSequencerClass, sequence_changed),
			gtk_marshal_NONE__NONE,
			GTK_TYPE_NONE,
			0);
	gtk_object_class_add_signals(object_class,
				     sequencer_signals,
				     LAST_SIGNAL);
}

GtkType gdam_sequencer_get_type()
{
	static GtkType sequencer_type = 0;
	if (!sequencer_type) {
		static const GtkTypeInfo sequencer_info = {
			"GdamSequencer",
			sizeof(GdamSequencer),
			sizeof(GdamSequencerClass),
			(GtkClassInitFunc) gdam_sequencer_class_init,
			(GtkObjectInitFunc) gdam_sequencer_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_PRODUCER;
		sequencer_type = gtk_type_unique(parent,
		                                 &sequencer_info);
	}
	return sequencer_type;
}

GdamSequencer* gdam_sequencer_new()
{
	GdamSequencer* seq = GDAM_SEQUENCER(gtk_type_new(GDAM_TYPE_SEQUENCER));
	return seq;
}

GdamSequencerElement*	gdam_sequencer_add_element(GdamSequencer* sequencer,
	                                           int            buffer_id,
						   gdouble        beat_offset,
	                                           gdouble        rate,
	                                           gdouble        gain,
	                                           int            sample_offset,
	                                           int            output_len)
{
	GdamSequencerElement* element;
	g_return_val_if_fail(buffer_id != 0, NULL);

	gdam_debug("gdam_sequencer_add_element: "
	           "id=%d, beat=%g, rate/gain=%g/%g; outputlen=%d",
		   buffer_id, beat_offset, rate, gain, output_len);

	gdam_buffer_manager_ref(buffer_id, GDAM_MODEL(sequencer));

	element = g_new(GdamSequencerElement, 1);
	element->buffer_id = buffer_id;
	element->beat_offset = beat_offset;
	element->rate = rate;
	element->gain = gain;
	element->sample_offset = sample_offset;
	element->output_length = output_len;
	sequencer->elements = g_list_prepend(sequencer->elements, element);
	if (sequencer->frozen == 0)
		gdam_sequencer_update_sequence(sequencer);
	return element;
}

typedef struct _ImmediateAddition ImmediateAddition;
struct _ImmediateAddition {
	GdamSequencerElement* element;
	GdamSequencer* sequencer;
};

static void successfully_got_time(int num_args, GdamArg* args, gpointer data)
{
	ImmediateAddition* ia = (ImmediateAddition*)data;
	gint64 position;
	gint64 sequence_length;
	gdouble beatnum;
	GdamSequencer* sequencer = ia->sequencer;
	position = GDAM_ARG_INT64(args);
	sequence_length = (int) beat_to_samples(sequencer,
						sequencer->num_beats);
	if (sequencer->immediate_lag > 0){
		gint64 pos = position;
		pos -= (gint64)(sequencer->immediate_lag * 44100.0);
		if (pos < 0){
			pos += sequence_length;
		}
		position = pos;
	}

	beatnum = samples_to_beats(sequencer, position);

	/* TODO: make this snapping optional?? */
	if (sequencer->num_sub_beats) {
		beatnum *= sequencer->num_sub_beats;
		beatnum += 0.5;
		beatnum = floor(beatnum);
		beatnum /= sequencer->num_sub_beats;
	}

	ia->element->beat_offset = beatnum;
	sequencer->elements = g_list_prepend(sequencer->elements, ia->element);
	if (sequencer->frozen == 0)
		gdam_sequencer_update_sequence(sequencer);
	g_free(ia);
}
static void didnt_successfully_get_time(gpointer data)
{
	ImmediateAddition* ia = (ImmediateAddition*)data;
	g_free(ia->element);
	g_free(ia);
}
GdamSequencerElement*	gdam_sequencer_add_element_immediate
                                                  (GdamSequencer* sequencer,
	                                           int            buffer_id,
	                                           gdouble        rate,
	                                           gdouble        gain,
	                                           int            sample_offset,
	                                           int            output_len)
{
	ImmediateAddition* ia = g_new(ImmediateAddition, 1);
	GdamSequencerElement* element = g_new(GdamSequencerElement, 1);
	GdamModel* model = GDAM_MODEL(sequencer);
	const char* arg_names[] = { "position" };

	gdam_debug("gdam_sequencer_add_element: "
	           "id=%d, beat=TBD, rate/gain=%g/%g; outputlen=%d",
		   buffer_id, rate, gain, output_len);

	gdam_buffer_manager_ref(buffer_id, GDAM_MODEL(sequencer));
	element->buffer_id = buffer_id;
	element->beat_offset = -1;
	element->rate = rate;
	element->gain = gain;
	element->sample_offset = sample_offset;
	element->output_length = output_len;
	ia->element = element;
	ia->sequencer = sequencer;
	if (model->source_id == 0) {
		gdouble stopped_position;
		stopped_position = GDAM_PRODUCER(sequencer)->stopped_position;
		ia->element->beat_offset = seconds_to_beats(sequencer,
		                                    stopped_position);
		sequencer->elements = g_list_prepend(sequencer->elements,
		                                     ia->element);
		return element;
	}
		
	gdam_channel_append_request(model->channel,
			gdam_request_get_args(model->source_id,
			                      1,
					      arg_names,
					      successfully_got_time,
					      didnt_successfully_get_time,
					      ia)
	);
	return element;
}

static void remove_model(GdamSequencerElement* element, GdamSequencer* seq)
{
	gdam_sequencer_remove_element(seq, element);
}

static void gdam_sequencer_detach_from_buffer(GdamModel*     model,
				              guint          buffer_id)
{
	GSList* kill_elements = NULL;
	GList* elements;
	GdamSequencer* sequencer = GDAM_SEQUENCER(model);
	for (elements = sequencer->elements;
	     elements != NULL;
	     elements = elements->next) {
	     	GdamSequencerElement* element = elements->data;
		if (element->buffer_id == buffer_id)
			kill_elements = g_slist_prepend(kill_elements, element);
	}
	if (kill_elements) {
		gdam_sequencer_freeze(sequencer);
		g_slist_foreach(kill_elements,
			        (GFunc) remove_model,
				sequencer);
		gdam_sequencer_thaw(sequencer);
		gtk_signal_emit(GTK_OBJECT(sequencer),
		                sequencer_signals[SEQUENCE_CHANGED]);
		g_slist_free(kill_elements);
	}
}

void gdam_sequencer_remove_element(GdamSequencer* sequencer,
				   GdamSequencerElement* element)
{
	sequencer->elements = g_list_remove(sequencer->elements, element);
	gdam_buffer_manager_unref(element->buffer_id,
	                          GDAM_MODEL(sequencer));
	g_free(element);
	gdam_sequencer_update_sequence(sequencer);
}

void gdam_sequencer_set_num_beats (GdamSequencer* sequencer,
                                   gdouble        num_beats)
{
	GdamModel* model = GDAM_MODEL(sequencer);
	int length = beat_to_samples(sequencer, num_beats);

	sequencer->num_beats = num_beats;

	gtk_signal_emit(GTK_OBJECT(sequencer),
		        sequencer_signals[SET_NUM_BEATS],
			num_beats);
	gtk_signal_emit_by_name(GTK_OBJECT(sequencer),
	                        "update_length",
			        (gdouble)length / 44100.0);

	if (model->source_id == 0)
		return;


	gdam_channel_append_request(model->channel,
		gdam_request_set_args(model->source_id,
		                      1,
				      gdam_arg_new_from_valist(
				      		1,
						"sequence_length",
						GDAM_ARG_TYPE_INT,
						length
				      ),
				      NULL,
				      NULL,
				      NULL));
}

void gdam_sequencer_freeze(GdamSequencer* sequencer)
{
	sequencer->frozen++;
}

void gdam_sequencer_thaw(GdamSequencer* sequencer)
{
	g_return_if_fail(sequencer->frozen > 0);
	if (--sequencer->frozen == 0)
		gdam_sequencer_update_sequence(sequencer);
}

static void clear_one(GdamSequencerElement* element, GdamModel* sequencer)
{
	gdam_buffer_manager_unref(element->buffer_id, sequencer);
	g_free(element);
}
void gdam_sequencer_clear(GdamSequencer* sequencer)
{
	g_list_foreach(sequencer->elements, (GFunc) clear_one, sequencer);
	g_list_free(sequencer->elements);
	sequencer->elements = NULL;
	gdam_sequencer_update_sequence(sequencer);
}

static void reference_element_buffer(GdamSequencerElement* element,
                                     GdamModel*            model)
{
	gdam_buffer_manager_ref(element->buffer_id, model);
}

void gdam_sequencer_set_elements(GdamSequencer* sequencer,
                                 GList*         elements)
{
	g_list_foreach(elements, (GFunc) reference_element_buffer, sequencer);
	gdam_sequencer_freeze(sequencer);
	gdam_sequencer_clear(sequencer);
	sequencer->elements = elements;
	gdam_sequencer_thaw(sequencer);
}

static void rescale_one(GdamSequencerElement* element, 
			GdamSequencerRescaleInfo* rescale_info)
{
	element->beat_offset *= rescale_info->rescale_factor;
}

/* Rescale the position of elements within a sequence
   by scale_factor.  TODO: incorporate set_length? */
void gdam_sequencer_rescale_elements(GdamSequencer* sequencer, 
				     gdouble	    scale_factor)
{
	GdamSequencerRescaleInfo rescale_info;
	rescale_info.sequencer = sequencer;
	rescale_info.rescale_factor = scale_factor;
	g_list_foreach(sequencer->elements, (GFunc)rescale_one, &rescale_info);
}

static gboolean gdam_sequencer_set_beat_info(GdamModel*           model,
	                                      const GdamBeatInfo* new_info)
{
	model->beat_info = *new_info;
	gdam_sequencer_update_sequence(GDAM_SEQUENCER(model));
	return TRUE;
}

/* Match beat info - adjusts length of sequence and position of
 * samples so that the tempo of the sequence matches the tempo
 * passed it.  As the sequence can only match beat info by setting
 * its real beat info, this is done.
 */
static gboolean gdam_sequencer_match_beat_info(GdamModel*	      model,
				  	       const GdamBeatInfo*    new_info)
{
	GdamSequencer* sequencer;
	sequencer = GDAM_SEQUENCER(model);
	gdam_sequencer_freeze(sequencer);
	gdam_model_set_beat_info(model, new_info);
	gdam_sequencer_set_num_beats(sequencer, sequencer->num_beats);
	gdam_sequencer_thaw(sequencer);
	return TRUE;
}
